/*
 * File......: A2DISK.PRG
 * Author....: Brian Dukes
 * BBS.......: The Dark Knight Returns
 * Net/Node..: 050/069
 * User Name.: Brian Dukes
 * Date......: 08/08/1995
 * Revision..: 1.0
 *
 * This is an original work by Brian Dukes and is placed in the public
 * domain.
 *
 * Modification history:
 * ---------------------
 *
 * $Log$
 *
 */

/*  $DOC$
 *  $FUNCNAME$
 *      GT_a2Disk()
 *  $CATEGORY$
 *      ARRAY
 *  $ONELINER$
 *      Save a complete array to disk (including nested arrays)
 *  $SYNTAX$
 *      GT_a2Disk( aArray, sFileName )
 *  $ARGUMENTS$
 *      aArray    - Array to save
 *      sFileName - name of file to store array to
 *  $RETURNS$
 *      .T. - Everything went OK!
 *      .F. - Whoops, problem creating file use FERROR() to retrieve error
 *  $DESCRIPTION$
 *      This function will save the complete contents of an array
 *      to a file, for later retrieval.  It will save Character,
 *      String, Logical, Date, Numeric, NIL and Array values.
 *
 *      So nested arrays are well catered for, and nested to any depth!
 *
 *      IT CANNOT SAVE CODEBLOCKS!!    THINK ABOUT IT!!
 *
 *      However, if it comes across a codeblock, or any other unknown
 *      data type that element will be saved as NIL
 *
 *      NOTE:  This function has a recursive portion to it,  so if you
 *             are trying to save very very deep arrays, you may need
 *             to increase your STACK (or PROCEDURE DEPTH)
 *
 *
 *  $EXAMPLES$
 *  function main()
 *     local a1 := {"element 1","this is element 2","number 3"}
 *     local a2 := {1,2,3,4,"hello world!",.f.,.t.,NIL,date(), ;
 *                  3.1459,'D',,"HI THERE",12.33,10.22}
 *     local a3 := {{1,2,3,4},{"a","b","c"},"hello", ;
 *                  {"subarray",{"and another",{"thats it!",.t.}}},;
 *                  date()}
 *     local a4 := NIL
 *     local a5 := NIL
 *     local a6 := NIL
 *
 *     GT_a2Disk(a1,"test1")
 *     GT_a2Disk(a2,"test2")
 *     GT_a2Disk(a3,"test3")
 *
 *     a4 := GT_disk2a("test1")
 *     a5 := GT_disk2a("test2")
 *     a6 := GT_disk2a("test3")
 *
 *     // Now using MRDEBUG Our Favourite Debugger, inspect the arrays
 *     // to check everything is OK!
 *  return NIL
 *
 *  $SEEALSO$
 *      GT_DISK2A()
 *  $INCLUDE$
 *
 *  $END$
 */



#define A_NIL        chr(0)
#define A_CHARACTER  chr(1)
#define A_STRING     chr(2)
#define A_NUMERIC    chr(3)
#define A_DATE       chr(4)
#define A_LOGICAL    chr(5)
#define A_ARRAY      chr(6)


#ifdef DEMO
function main()
   local a1 := {"element 1","this is element 2","number 3"}
   local a2 := {1,2,3,4,"hello world!",.f.,.t.,NIL,date(), ;
                3.1459,'D',,"HI THERE",12.33,10.22}
   local a3 := {{1,2,3,4},{"a","b","c"},"hello", ;
                {"subarray",{"and another",{"thats it!",.t.}}},;
                date()}
   local a4 := NIL
   local a5 := NIL
   local a6 := NIL

   GT_a2Disk(a1,"test1")
   GT_a2Disk(a2,"test2")
   GT_a2Disk(a3,"test3")

   a4 := GT_disk2a("test1")
   a5 := GT_disk2a("test2")
   a6 := GT_disk2a("test3")

   // Now using MRDEBUG Our Favourite Debugger, inspect the arrays
   // to check everything is OK!
return NIL
#endif




//
// To Save an array to disk, we need to know some things about each element in the array
// thus,  marker characters will be used to denote type.
//
// Marker  chr(0) == NIL
//         chr(1) == CHARACTER
//         chr(2) == STRING
//         chr(3) == NUMERIC
//         chr(4) == DATE
//         chr(5) == LOGICAL
//         chr(6) == ARRAY
//
// Optionally the character following a marker character will denote something special!
//
// Marker chr(0):NIL       -> No additional characters, the next character is a marker for the next element!
//        chr(1):CHARACTER -> The Next character is the value of this element
//        chr(2):STRING    -> The Next TWO binary digits denote the length of the string!
//        chr(3):NUMERIC   -> The Next TWO binary digits describe the length of the stringized number!
//        chr(4):DATE      -> No additional info
//        chr(5):LOGICAL   -> No additional info
//        chr(6):ARRAY     -> The Next TWO binary digits denote the number of elements following for this array.
//

function GT_a2Disk(a, fname)
   local nHandle := fcreate(fname,0)
   local lStatus := .T.

   IF nHandle == -1
      lStatus := .F.
   ELSE
      fwrite(nHandle,A_ARRAY+i2Bin(len(a)),3)
      _a2dRecurs(a,nHandle)
      fclose(nHandle)
   ENDIF
return .T.


//  This function will be called recursively, when it needs to navigate a sub-array

function _a2dRecurs(a,nHandle)
   local nLen := len(a)
   local n    := 0
   local t    := ""
   local l    := 0
   local d    := 0

   FOR n := 1 TO nLen
      t := valtype(a[n])
      DO CASE
         CASE t == "C"     // CHARACTER!
            l := len(a[n])
            IF l == 1
               fwrite(nHandle,A_CHARACTER+a[n],2)
            ELSE
               fwrite(nHandle,A_STRING+i2Bin(l)+a[n],3+l)
            ENDIF
         CASE t == "N"     // NUMERIC
            l := len(str(a[n]))
            fwrite(nHandle,A_NUMERIC+i2Bin(l)+str(a[n]),3+l)
         CASE t == "D"     // DATE
            fwrite(nHandle,A_DATE+dtos(a[n]),9)
         CASE t == "L"     // LOGICAL
            fwrite(nHandle,A_LOGICAL+if(a[n],"T","F"),2)
         CASE t == "A"     // ARRAY ... Lets Recurse!!!
            fwrite(nHandle,A_ARRAY+i2Bin(len(a[n])),3)
            _a2dRecurs(a[n],nHandle)
         OTHERWISE         // Treat as NIL, this includes codeblocks and other undefined types!
            fwrite(nHandle,A_NIL,1)
      ENDCASE
   NEXT
return NIL




