/*
 * File......: DISK2A.PRG
 * Author....: Brian Dukes
 * BBS.......: The Dark Knight Returns
 * Net/Node..: 050/069
 * User Name.: Brian Dukes
 * Date......: 08/08/1995
 * Revision..: 1.0
 *
 * This is an original work by Brian Dukes and is placed in the public
 * domain.
 *
 * Modification history:
 * ---------------------
 *
 * $Log$
 *
 */

/*  $DOC$
 *  $FUNCNAME$
 *      GT_Disk2A()
 *  $CATEGORY$
 *      ARRAY
 *  $ONELINER$
 *      Restore a complete array to from disk (including nested arrays)
 *  $SYNTAX$
 *      GT_Disk2A( sFileName )
 *  $ARGUMENTS$
 *      sFileName - name of file to retrieve array from
 *  $RETURNS$
 *      aArray - Returns the array!
 *  $DESCRIPTION$
 *      This function is the counter-part function to GT_a2Disk()
 *      and this function restores completely the array that was
 *      saved to disk!
 *
 *      IT CANNOT RESTORE CODEBLOCKS!!    THINK ABOUT IT!!
 *
 *      However, if it comes across a codeblock, or any other unknown
 *      data type that element will be saved as NIL
 *
 *      NOTE:  This function has a recursive portion to it,  so if you
 *             are trying to save very very deep arrays, you may need
 *             to increase your STACK (or PROCEDURE DEPTH)
 *
 *
 *  $EXAMPLES$
 *  function main()
 *     local a1 := {"element 1","this is element 2","number 3"}
 *     local a2 := {1,2,3,4,"hello world!",.f.,.t.,NIL,date(), ;
 *                  3.1459,'D',,"HI THERE",12.33,10.22}
 *     local a3 := {{1,2,3,4},{"a","b","c"},"hello", ;
 *                  {"subarray",{"and another",{"thats it!",.t.}}},;
 *                  date()}
 *     local a4 := NIL
 *     local a5 := NIL
 *     local a6 := NIL
 *
 *     GT_a2Disk(a1,"test1")
 *     GT_a2Disk(a2,"test2")
 *     GT_a2Disk(a3,"test3")
 *
 *     a4 := GT_disk2a("test1")
 *     a5 := GT_disk2a("test2")
 *     a6 := GT_disk2a("test3")
 *
 *     // Now using MRDEBUG Our Favourite Debugger, inspect the arrays
 *     // to check everything is OK!
 *  return NIL
 *
 *  $SEEALSO$
 *      GT_A2DISK()
 *  $INCLUDE$
 *
 *  $END$
 */



#define A_NIL        chr(0)
#define A_CHARACTER  chr(1)
#define A_STRING     chr(2)
#define A_NUMERIC    chr(3)
#define A_DATE       chr(4)
#define A_LOGICAL    chr(5)
#define A_ARRAY      chr(6)


function GT_disk2a(fname)
   local a         := {}
   local nHandle   := fopen(fname,0)
   local lStatus   := .T.
   local buf       := space(255)
   local nElements := 0

   IF nHandle == -1
      a := NIL
   ELSE
      fread(nHandle,@buf,3)
      IF left(buf,1) == A_ARRAY
         nElements := bin2i(substr(buf,2,2))
         a := _d2aRecurs(a,nHandle,nElements)
      ENDIF
      fclose(nHandle)
   ENDIF
return a



function _d2aRecurs(aOrg,nHandle,nElements)
   local n   := 0
   local t   := " "
   local buf := space(255)
   local l   := 0
   local a   := aClone(aOrg)

   FOR n := 1 TO nElements
      fread(nHandle,@t,1)
      DO CASE
         CASE t == A_NIL
            aadd(a,NIL)
         CASE t == A_CHARACTER
            fread(nHandle,@buf,1)
            aadd(a,left(buf,1))
         CASE t == A_STRING
            fread(nHandle,@buf,2)
            l := bin2i(buf)
            fread(nHandle,@buf,l)
            aadd(a,left(buf,l))
         CASE t == A_NUMERIC
            fread(nHandle,@buf,2)
            l := bin2i(buf)
            fread(nHandle,@buf,l)
            aadd(a,val(left(buf,l)))
         CASE t == A_DATE
            fread(nHandle,@buf,8)
            aadd(a,ctod(substr(buf,7,2)+"/"+substr(buf,5,2)+"/"+left(buf,4)))
         CASE t == A_LOGICAL
            fread(nHandle,@buf,1)
            aadd(a,if(left(buf,1)=="T",.t.,.f.))
         CASE t == A_ARRAY
            fread(nHandle,@buf,2)
            l := bin2i(buf)
            aadd(a, {})
            a[len(a)] := _d2aRecurs(a[len(a)],nHandle,l)
      ENDCASE
   NEXT
return a
