#define HIGH_PRIME   30296

#define RND_SEEDED   1
#define RND_RANDOM   2

/*  $DOC$
 *  $FUNCNAME$
 *      GT_Rand1()
 *  $CATEGORY$
 *      Misc
 *  $ONELINER$
 *      Return a random number up to a max value.
 *  $SYNTAX$
 *      GT_Rand1( <nMax> ) --> nRandom
 *  $ARGUMENTS$
 *      <nMax> is the max value for the random number. The number generated
 *      will be between 1 and <nMax> inclusive.
 *  $RETURNS$
 *      A random number of sorts.
 *  $DESCRIPTION$
 *      This function is a Clipper version of a function that appeared in
 *      a article in issue 41 of Computer Shopper (July 1991) on page 198.
 *
 *      It will return a random number of sorts.
 *  $SEEALSO$
 *
 *  $END$
 */

Function GT_Rand1( nMax )
   Static aRand   := { .F., 0 }
   Local  nRandom := 0

   If !aRand[ RND_SEEDED ]
      aRand[ RND_RANDOM ] := ( seconds() * 100 )
      aRand[ RND_SEEDED ] := .T.
   EndIf

   aRand[ RND_RANDOM ] := ( ( 177 * aRand[ RND_RANDOM ] ) % HIGH_PRIME )

   nRandom := min( int( nMax * ( aRand[ RND_RANDOM ] / HIGH_PRIME ) ) + 1, nMax )

Return( nRandom )
