#include "error.ch"
#INCLUDE "inkey.CH"

#define NUMSHAPES 7
#DEFINE SHOWSHAPE    .T. 
#DEFINE ERASESHAPE   .F. 
#DEFINE HI_SCORE_FILE   "TETRIS.HI"
#DEFINE _MAX_HISCORES   5

#Ifdef DEMO
   function main()
      GT_Tetris(.t.,.t.)
   return NIL
#ENDIF

function GT_Tetris(lSound,lHiScoreTable)
   local lStillgoing := .T. 
   local nClr        := setcolor()
   local cOrgScreen  := savescreen(0,0,maxrow(),maxcol())
   local nOrgRows    := maxrow()
   local nOrgCols    := maxcol()
   local nOrgCursor  := setcursor(0)

   setcolor("W/N")
   clear screen
   setmode(25,40)

   DO WHILE lStillGoing
      lStillGoing := _Tetris(lSound,lHiScoreTable)
   ENDDO 

   setmode(nOrgRows+1,nOrgCols+1)
   restscreen(0,0,nOrgRows,nOrgCols,cOrgScreen)
   setcursor(nOrgCursor)
   setcolor(nClr)
return NIL


function _Tetris(lSound,lHiScoreTable)
   local lStillGoing := .T. 
   local lGameOver   := .F. 
   local nKey        := 0
   local nConst      := _GetTimeConstant()
   local aShapes     := _ShapeTable()
   local nRow        := 3
   local nCol        := 10
   local nShape      := int(GT_rand1(NUMSHAPES))
   local nWay        := 1
   local nNewCol     := 1
   local nNewRow     := 1
   local nNewWay     := 1
   local aTable      := {}
   local nScore      := 0
   local nNextShape  := int(GT_rand1(NUMSHAPES))
   local aHiScores   := {}
   local n           := 0
   local i           := 0

   IF lSound == NIL
      lSound := .T.     // Default to SOUND:ON
   ENDIF
   IF lHiScoreTable == NIL
      lHiScoreTable := .T. 
   ENDIF 
    
   _isNoisy(lSound)     // Set Default value

   aTable := _TetrisScreen()
   _isNoisy()     // Show status
   _ShowScore(nScore)
   _ShowLinesCompleted(0)
   IF lHiScoreTable
      aHiScores := _ShowHiScores()
   ENDIF 
   _ShowNextShape(nNextShape,aShapes,SHOWSHAPE)
   _DrawShape(nRow,nCol,SHOWSHAPE,aShapes,nShape,nWay,aTable)

   DO WHILE lStillGoing
      for n := 1 to nConst
         nKey := inkey()
         nNewCol := nCol
         nNewRow := nRow
         nNewWay := nWay
         if nShape != 0 .and. nKey != 0
            dispbegin()
            _DrawShape(nRow,nCol,ERASESHAPE,aShapes,nShape,nWay,aTable)
            do case
               case nKey == K_ESC
                  n := nConst
               case nKey == asc("q") .OR. nKey == asc("Q")
                  _isNoisy(!_isNoisy())

               case nKey == 52 .OR. nKey == K_LEFT  // numeric 4/LEFT
                  if _OKToMove(K_LEFT,nRow,nCol,aShapes,nShape,nWay,aTable)
                     nNewCol--
                  endif

               case nKey == 54 .OR. nKey == K_RIGHT  // numeric 6/RIGHT
                  if _OKToMove(K_RIGHT,nRow,nCol,aShapes,nShape,nWay,aTable)
                     nNewCol++
                  endif

               case nKey == 53 .OR. nKey == 32 .OR. nKey == K_UP  // numeric 5/SPACE/UP
                  if _OKToMove(K_UP,nRow,nCol,aShapes,nShape,iif(nWay == 4, 1, nWay+1) ,aTable)
                     nNewWay := iif(nWay == 4, 1, nWay+1)
                  endif

               case nKey == 50 .OR. nKey == K_DOWN       // numeric 2/DOWN
                  dispend()
                  DO WHILE _OKToMove(K_DOWN,nRow,nCol,aShapes,nShape,nWay,aTable)
                     _DrawShape(++nRow,nCol,SHOWSHAPE,aShapes,nShape,nWay,aTable)
                     FOR i := 1 TO 1+(nConst/5)    // provid a delay of sorts!
                        inkey()
                     NEXT 
                     _DrawShape(nRow,nCol,ERASESHAPE,aShapes,nShape,nWay,aTable)
                     nNewRow := nRow
                  ENDDO 
                  dispbegin()
            endcase
            nRow := nNewRow
            nCol := nNewCol
            nWay := nNewWay
            _DrawShape(nRow,nCol,SHOWSHAPE,aShapes,nShape,nWay,aTable)
            dispend()
         endif
      next

      DO WHILE dispcount() > 0      // just in case!
         dispend()
      ENDDO 

  //    showtable(aTable)
      dispbegin()
      _DrawShape(nRow,nCol,ERASESHAPE,aShapes,nShape,nWay,aTable)
      if _OKToMove(K_DOWN,nRow,nCol,aShapes,nShape,nWay ,aTable)
         _DrawShape(++nRow,nCol,SHOWSHAPE,aShapes,nShape,nWay,aTable)
         // ok slowly get quicker
         nConst -= 2
      ELSE
         _DrawShape(nRow,nCol,SHOWSHAPE,aShapes,nShape,nWay,aTable)
         nScore := _CheckOffLines(aTable,nScore)
         nRow := 3
         nCol := 10
         nWay := 1
         _ShowNextShape(nNextShape,aShapes,ERASESHAPE)
         nShape := nNextShape
         nNextShape := int(GT_rand1(NUMSHAPES))
         _ShowNextShape(nNextShape,aShapes,SHOWSHAPE)
         // detect game over here!
         if .NOT. _OKToMove(K_DOWN,nRow,nCol,aShapes,nShape,nWay ,aTable)
            lGameOver := .T. 
            lStillGoing := .F. 
         endif
      endif
      dispend()

      IF nKey == K_ESC
         lStillGoing := .F. 
      ENDIF
   ENDDO 

   IF lGameOver
      lStillGoing := _GameOver(nScore,lHiScoreTable,aHiScores)
   ENDIF 
return lStillGoing



static function _TetrisScreen()
   local nWidth  := 21
   local nLength := 24
   local n       := 0
   local i       := 0
   local aTable  := {}

   dispbegin()
   setcolor("W/N")
   clear screen
   dispbox(0,1,24,21,"  ","G/N")
   dispend()

   FOR n := 1 TO nLength-1
      aadd(aTable,{})
      FOR i := 1 TO nWidth
         aadd(aTable[n],0)
      NEXT 
      aTable[n][1]      := 254    // Setup sides
      aTable[n][nwidth] := 254
   NEXT 
   n := nLength 
   aadd(aTable,{})
   FOR i := 1 TO nWidth
       aadd(aTable[n],254)
   NEXT 
return aTable



static function _DrawShape(nRow,nCol,lShowShape,aShape,nShape,nWay,aTable)
   local oldcolor := iif(lShowShape, setcolor('0/'+str(nShape)), setcolor())
   local cChar    := iif(lShowShape, 'X', ' ')
   local i        := 0
   local nX       := 0
   local nY       := 0

   for i :=0 to 3
      nX := nRow + aShape[nShape][nWay][(i*2)+1]
      nY := nCol + aShape[nShape][nWay][(i*2)+2]
      aTable[nX][nY] := if(lShowShape,nShape,0)
      DevPos(nX,nY)
      DevOut(cChar)
   next
   setcolor(oldcolor)
return nil



static function _ShapeTable()
   local aShapes     := { { {  0, -1,  0,  0,  0,  1,  0,  2},;
                            { -2,  0, -1,  0,  0,  0,  1,  0},;
                            {  0, -1,  0,  0,  0,  1,  0,  2},;
                            { -2,  0, -1,  0,  0,  0,  1,  0} ;
                          },;
                          { { -1,  0, -1,  1,  0,  0,  0,  1},;
                            { -1,  0, -1,  1,  0,  0,  0,  1},;
                            { -1,  0, -1,  1,  0,  0,  0,  1},;
                            { -1,  0, -1,  1,  0,  0,  0,  1} ;
                          },;
                          { {  0, -1,  0,  0,  0,  1,  1,  0},;
                            { -1,  0,  0,  0,  0,  1,  1,  0},;
                            { -1,  0,  0, -1,  0,  0,  0,  1},;
                            { -1,  0,  0, -1,  0,  0,  1,  0} ;
                          },;
                          { { -1, -1, -1,  0,  0,  0,  0,  1},;
                            { -1,  1,  0,  0,  0,  1,  1,  0},;
                            { -1, -1, -1,  0,  0,  0,  0,  1},;
                            { -1,  1,  0,  0,  0,  1,  1,  0} ;
                          },;
                          { {  0,  0,  0,  1,  1, -1,  1,  0},;
                            { -1,  0,  0,  0,  0,  1,  1,  1},;
                            {  0,  0,  0,  1,  1, -1,  1,  0},;
                            { -1,  0,  0,  0,  0,  1,  1,  1} ;
                          },;
                          { { -1, -1,  0, -1,  0,  0,  0,  1},;
                            { -1,  0,  0,  0,  1, -1,  1,  0},;
                            {  0, -1,  0,  0,  0,  1,  1,  1},;
                            { -1,  0, -1,  1,  0,  0,  1,  0} ;
                          },;
                          { { -1,  1,  0, -1,  0,  0,  0,  1},;
                            { -1, -1, -1,  0,  0,  0,  1,  0},;
                            {  0, -1,  0,  0,  0,  1,  1, -1},;
                            { -1,  0,  0,  0,  1,  0,  1,  1} ;
                          } ;
                        }
return aShapes



static function _GetTimeConstant()
   local nStartTime := 0
   local nTimeConst := 0

   nStartTime := seconds()
   do while seconds()-nStartTime <= 1
      nTimeConst++
   enddo
return nTimeConst



static function ArrErr(objErr)
    break objErr
return nil



static function _OKToMove(nDir,nRow,nCol,aShape,nShape,nWay,aTable)
   local i       := 1
   local lRet    := .T. 
   local nNewRow := if(nDir == K_DOWN, nRow+1, nRow)
   local nNewCol := if(nDir == K_LEFT, nCol-1, if(nDir == K_RIGHT, nCol+1, nCol))
   local nX      := 0
   local nY      := 0
   local bOldErr := ErrorBlock({|errObj| ArrErr(errObj)}) /*install my error handler*/
   local objErr  := NIL


   FOR i := 0 TO 3
      nX := nNewRow + aShape[nShape][nWay][(i*2)+1]
      nY := nNewCol + aShape[nShape][nWay][(i*2)+2]
      begin sequence
         IF aTable[nX][nY] != 0
            lRet := .f.
         endif
      recover using objErr
         if objErr:genCode == EG_BOUND   /*if out of bounds return false*/
            lRet := .f.
         else
            eval(bOldErr, objErr)     /*else chain to other error handler*/
         endif
      end sequence
   NEXT 

   errorblock(bOldErr)     /*reintall old handler*/
return lRet


static function showtable(aTable)
   local l := len(aTable)
   local lx := len(aTable[1])
   local i := 0
   local n := 0

   dispbegin()
   FOR i := 1 TO l
      devpos(i,25)
      FOR n := 1 TO lx
         ?? if(aTable[i][n]==0,' ',chr(aTable[i][n]))
      NEXT 
   NEXT 
   dispend()
return NIL


static function _CheckOffLines(aTable,nScore)
   local l         := len(aTable) - 1
   local lx        := len(aTable[1])
   local i         := 0
   local n         := 0
   local lComplete := .F. 

   FOR i := 1 TO l
      lComplete := .T. 
      FOR n := 1 TO lx
         IF aTable[i][n] == 0
            lComplete := .F. 
            EXIT
         ENDIF 
      NEXT 
      IF lComplete 
         IF _BonusRow(i,aTable)
            nScore := _BonusSound(nScore, _RowScore(i,aTable) * 10)
         ELSE
            nScore := _RowCompleteSound(nScore, _RowScore(i,aTable) )
         ENDIF 
         aTable := _RemoveLine(i,aTable)
         _ShowLinesCompleted(+1)
      ENDIF 
   NEXT 
return nScore


static function _RemoveLine(nRemove,aTable)
   local lx   := len(aTable[1])
   local i    := 0
   local cScr := ""

   aDel(aTable,nRemove)
   aINS(aTable,1)
   aTable[1] := {}

   FOR i := 1 TO lx
      aadd(aTable[1],0)
   NEXT 
   aTable[1][1]  := 254    // Setup sides
   aTable[1][lx] := 254
   // now need to repaint the screen!
   cScr := savescreen(0,0,nRemove-1,lx+1)
   restscreen(1,0,nRemove,lx+1,cScr)
return aTable



// Run across a completed row, and determine if its filled with shapes of the same colour
// this adds as a bonus
static function _Bonusrow(nRow,aTable)
   local lx   := len(aTable[1])
   local i    := 0
   local lRet := .T. 
   local nLast := aTable[nRow][2]

   FOR i := 3 TO lx-1
      IF aTable[nRow][i] != nLast
         lRet := .F. 
         EXIT
      ENDIF 
      nLast := aTable[nRow][i]
   NEXT 
return lRet


// Simple scoring mechanism, just add up the total of the rows!
static function _RowScore(nRow,aTable)
   local lx   := len(aTable[1])
   local i    := 0
   local nScore := 0

   FOR i := 2 TO lx-1
      nScore += aTable[nRow][i]
   NEXT 
return nScore


static function _RowCompleteSound(nScore,nAdd)
   local nNewScore := nScore
   local i         := 0
   local nDisps    := dispcount()

   DO WHILE dispcount() >0
      dispend()
   ENDDO 

   FOR i := 0 TO nAdd step 10
      nNewScore += 10
      _showScore(nNewScore)
      tone(if(_isNoisy(),130+(i*1.5),0),1)
   NEXT 

   FOR i := 1 TO ndisps
      dispbegin()
   NEXT 
return nNewScore



static function _BonusSound(nScore,nAdd)
   local nNewScore := nScore
   local i         := 0
   local nDisps    := dispcount()

   DO WHILE dispcount() >0
      dispend()
   ENDDO 

   @ 10,30 say "*BONUS*" colour "R+/N*"

   FOR i := 1 TO nAdd step 25
      nNewScore += 25
      _showScore(nNewScore)
      tone(if(_isNoisy(),130+(i*1.5),0),1)
   NEXT 

   @ 10,30 say "       " colour "N/N"

   FOR i := 1 TO ndisps
      dispbegin()
   NEXT 
return nNewScore


static function _showScore(nScore)
   @ 0, 25 say "SCORE" colour "R/N"
   @ 0, 31 say strzero(nScore,8) colour "BG+/N"
return NIL


static function _ShowNextShape(nShape,aShape,lShowShape)
   local oldcolor := iif(lShowShape, setcolor('0/'+str(nShape)), setcolor())
   local cChar    := iif(lShowShape, 'X', ' ')
   local i        := 0
   local nX       := 0
   local nY       := 0
   local nWay     := 1

   for i :=0 to 3
      nX := aShape[nShape][nWay][(i*2)+1]
      nY := aShape[nShape][nWay][(i*2)+2]
      DevPos(4+nX,25+nY)
      DevOut(cChar)
   next

   @ 3, 30 say "NEXT"  colour"GR+/N"
   @ 4, 30 say "SHAPE" colour"GR+/N"

   setcolor(oldcolor)
return NIL


static function _ShowLinesCompleted(nAdd)
   static nLines := 0

   IF nAdd == NIL
      nAdd := 0
   ENDIF 

   nLines += nAdd

   @ 1, 25 say "LINES" colour "R/N"
   @ 1, 31 say strzero(nLines,8) colour "BG+/N"
return nLines


static function _isNoisy(lSetState)
   static lSound := .T. 

   IF lSetState != NIL
      lSound := lSetState
   ENDIF 
   @ maxrow(),maxcol()-9 say "SOUND "+if(lSound,"ON ","OFF") colour "W/N"
return lSound


static function _ErrorBeep()
   local lSound := _isNoisy()
   
   tone(if(lSound,300,0),1)
   tone(if(lSound,200,0),1)
   tone(if(lSound,100,0),1)
   tone(if(lSound,300,0),1)
   tone(if(lSound,200,0),1)
   tone(if(lSound,100,0),1)
return NIL


function _OKbeep()
   local lSound := _isNoisy()

   tone(if(lSound,100,0),1)
   tone(if(lSound,200,0),1)
   tone(if(lSound,300,0),1)
   tone(if(lSound,400,0),1)
   tone(if(lSound,500,0),1)
   tone(if(lSound,600,0),1)
return NIL


static function _GameOver(nScore,lHiScores,aHiScores)
   local nDisps    := dispcount()
   local lContinue := .T. 
   local lGoing    := .T. 
   local nKey      := 0
   local n         := 0

   memvar GETLIST

   DO WHILE dispcount() >0
      dispend()
   ENDDO 
   @ 10,25 say "GAME OVER" colour "R+/N*"
   _ErrorBeep() // this forms a delay of sorts as well!
   IF lHiScores
      FOR n := 1 TO _MAX_HISCORES
         IF nScore > aHiScores[n][1]
            ADEL(aHiScores,len(aHiScores))
            AINS(aHiScores,n)
            aHiScores[n] := { nScore, "      " }
            GT_a2Disk(aHiScores,HI_SCORE_FILE)
            aHiScores := _ShowHiScores()
            @ 10,25 say "HIGH SCORE" colour "GR+/N*"
            _OKBeep()
            setcursor(3)
            @ 14+n, 32 get aHiScores[n][2] picture "!!!!!!" colour "W+/N,W+/N"
            READ
            setcursor(0)
            GT_a2Disk(aHiScores,HI_SCORE_FILE)
            aHiScores := _ShowHiScores()
            EXIT
         ENDIF 
      NEXT 
   ENDIF 
   tone(0,8)   // 1/2 second delay
   @ 10,24 say "PLAY AGAIN?"

   DO WHILE lGoing
      nKey := inkey(0)
      DO CASE
         CASE nKey == K_ESC .OR. nKey == asc("N") .OR. nKey == asc("n")
            lGoing    := .F. 
            lContinue := .F. 
         CASE nKey == asc("Y") .OR. nKey == asc("y")
            lGoing    := .F. 
            lContinue := .T. 
      ENDCASE 
   ENDDO 
return lContinue


static function _ShowHiScores()
   local aScore := {}
   local i      := 1

   IF file(HI_SCORE_FILE)
      aScore := GT_Disk2a(HI_SCORE_FILE)
   else
      FOR i := 1 TO _MAX_HISCORES
//         aadd(aScore,{5000-(i*1000),"      "})
         aadd(aScore,{5-(i),"      "})
      NEXT 

      IF _MAX_HISCORES >= 1
         aScore[1][2] :="TETRIS"
      ENDIF 
      IF _MAX_HISCORES >= 2
         aScore[2][2] :="    BY"
      ENDIF 
      IF _MAX_HISCORES >= 3
         aScore[3][2] :=" BRIAN"
      ENDIF 
      IF _MAX_HISCORES >= 4
         aScore[4][2] :=" DUKES"
      ENDIF 
      IF _MAX_HISCORES >= 5
         aScore[5][2] :="  1996"
      ENDIF 
   ENDIF 

   @ 13, 25 say "HALL OF FAME"
   FOR i := 1 TO _MAX_HISCORES
      @ 14+i, 23 say strzero(aScore[i][1],8)+" "+aScore[i][2] colour "R+/N"
   NEXT 
return aScore
